#ifndef __ConQueue__
#define __ConQueue__
#include <queue>
#include <mutex>
#include <condition_variable>

template <typename Data>
class ConcurrentQueue
{
private:
    std::queue<Data> queue;
    std::mutex queueMutex;
    std::condition_variable queueCv;

public:
    void push(Data const &data)
    {
        std::unique_lock<std::mutex> lock(queueMutex);
        queue.push(data);
        lock.unlock();
        queueCv.notify_one();
    }

    bool empty()
    {
        std::unique_lock<std::mutex> lock(queueMutex);
        return queue.empty();
    }

    bool try_pop(Data &popped_value)
    {
        std::unique_lock<std::mutex> lock(queueMutex);
        if (queue.empty())
        {
            return false;
        }

        popped_value = queue.front();
        queue.pop();
        return true;
    }

    void wait_and_pop(Data &popped_value)
    {
        std::unique_lock<std::mutex> lock(queueMutex);
        queueCv.wait(lock, [&]{return queue.size() > 0;});
        popped_value = queue.front();
        queue.pop();
    }
};
#endif