#include <rdma/rdma_cma.h>
#include <rdma/rdma_verbs.h>
#include <stdint.h>
#include <thread>
#include <netdb.h>
#include <arpa/inet.h>
#include <map>
#include <vector>
#include <mutex>
#include <shared_mutex>

#ifndef __RDMASERVERENDPOINTGROUP__
#define __RDMASERVERENDPOINTGROUP__

#include "RdmaSalEndpoint.hpp"
#include "CqEventData.hpp"
#include "Executor.hpp"
#include "RdmaSalCqProcessor.hpp"

class RdmaServerEndpointGroup
{
    /*
    * Variables to maintain Group state
    */
    static int CONN_STATE_INITIALIZED;
    static int CONN_STATE_BINDED;
    static int CONN_STATE_CONNECTED;
    static int CONN_STATE_CLOSED;

    bool _stop{false};
    std::thread *_cmEventThread{NULL};
    struct rdma_event_channel *_eventChannel{NULL};
    Executor *_executor;
    

    struct rdma_cm_id *_cm_id{NULL};
    RdmaSalCqProcessor *_salCqProcessor{NULL};
    std::vector<RdmaSalEndpoint *> *_salEps{NULL};
    std::unordered_map<uint32_t, RdmaSalEndpoint *> *_qpSalEndpointMap{NULL};

    
    /*
    * variables to maintain queue state
    */
    int _sendQueueSize{0};
    int _recvQueueSize{0};
    int _compQueueSize{0};
    int _sendMsgSize{0};
    int _recvMsgSize{0};
    rocksdb::DB *_db;

    mutable std::shared_mutex _salMutex;

    void clientClose();

public:
    RdmaServerEndpointGroup(int sendQueueSize, int recvQueueSize, int compQueueSize, int sendMsgSize, int recvMsgSize);
    void bind(const char *ip, const char *port, int backlog);

    struct ibv_cq *createSalCq(struct rdma_cm_id *id);
    void dispatchSalCqEvents(ibv_wc wc[], int size);
    void createExecutor(int threadSize);

    /* 
    * Sending false will run cmProcessor on calling thread used to save creation of new thread
    *  when main thread can also do same work
    */
    void startCmProcessor(bool newThread);
    void processCmEvents();
    void processCmEvent(struct rdma_cm_event *event);
    void createEpCmEvent(struct rdma_cm_event *event);
    void close();
};

#endif