#include "RdmaClientEndpointGroup.hpp"

RdmaClientEndpointGroup::RdmaClientEndpointGroup(int sendQueueSize, int recvQueueSize, int compQueueSize, int sendMsgSize,
                                                 int recvMsgSize,int maxInLine, int timeout)
    : _sendQueueSize(sendQueueSize), _recvQueueSize(recvQueueSize), _compQueueSize(compQueueSize),
      _sendMsgSize(sendMsgSize), _recvMsgSize(recvMsgSize), _maxInLine(maxInLine),_timeoutMs(timeout)
{
     std::cout << "RdmaClientEndpointGroup : Step 1 creating event channel" << std::endl;
    _eventChannel = rdma_create_event_channel();
    _stop = false;
    if (_eventChannel == NULL)
    {
        std::cout << "RdmaClientEndpointGroup : error creating event channel";
    }
}

void RdmaClientEndpointGroup::start()
{
    _cmEventThread = new std::thread(&RdmaClientEndpointGroup::processCmEvents, this);
    pthread_setname_np(_cmEventThread->native_handle(),"ClientCMProcessor");
}

void RdmaClientEndpointGroup::processCmEvents()
{
    int ret;
    struct rdma_cm_event *event;
    std::cout << "RdmaClientEndpointGroup : starting cm processing thread" << std::endl;
    while (!_stop)
    {
        ret = rdma_get_cm_event(_eventChannel, &event);
        if (ret)
        {
            std::cout << "RdmaClientEndpointGroup : rdma_get_cm_event failed" << std::endl;
            continue;
        }
        processCmEvent(event);
        ret = rdma_ack_cm_event(event);
        if (ret)
        {
            std::cout << "RdmaClientEndpointGroup : rdma_ack_cm_event failed";
        }
    }
}

void RdmaClientEndpointGroup::processCmEvent(struct rdma_cm_event *event)
{
    std::cout << "RdmaClientEndpointGroup :  event" << rdma_event_str(event->event) << std::endl;
    if (event->id != NULL && event->id->context != NULL)
    {
        if (event->event == RDMA_CM_EVENT_ADDR_RESOLVED && event->id != NULL)
        {
            ((RdmaClientEndpoint *)event->id->context)->createResources(createCq(event->id));
        }
        ((RdmaClientEndpoint *)event->id->context)->processCmEvent(event);
        if(event->event == RDMA_CM_EVENT_ADDR_RESOLVED)
        {
            _cqProcessor->registerEp(event->id->qp->qp_num,((RdmaClientEndpoint *)event->id->context));
        }
        
    }
    else
    {
        std::cout << "RdmaClientEndpointGroup : Not able to procces CM EVent";
        std::cout << rdma_event_str(event->event) << event->id << " ";
        std::cout << event->listen_id << std::endl;
    }
}

RdmaClientEndpoint *RdmaClientEndpointGroup::createEndpoint()
{
    struct rdma_cm_id *id = NULL;
    int ret = rdma_create_id(_eventChannel, &id, NULL, RDMA_PS_TCP);
    std::cout<<"id "<<id<<std::endl;
    if (ret == -1)
        std::cout << "CMProcesor : rdma_create_id failed" << std::endl;
    RdmaClientEndpoint *endpoint = new RdmaClientEndpoint(id,
                                                          _sendQueueSize, _recvQueueSize,
                                                          _sendMsgSize, _recvMsgSize,_maxInLine, _timeoutMs);
    id->context = (void *)endpoint;
    return endpoint;
}

struct ibv_cq *RdmaClientEndpointGroup::createCq(struct rdma_cm_id *id)
{
    if (_cqProcessor == NULL)
    {
        std::cout << "RdmaClientEndpointGroup : Creating CQ processor" << std::endl;
        _cqProcessor = new RdmaCqProcessor(id->verbs, _compQueueSize);
        _cqProcessor->start();
    }
    return _cqProcessor->getCq();
}

void RdmaClientEndpointGroup::close()
{
    _stop = true;
    _cmEventThread->join();
    rdma_destroy_event_channel(_eventChannel);
}