#include "TaskThread.hpp"
#include "MessageFormats.hpp"
TaskThread::TaskThread(int id, int cpu, ConcurrentQueue<struct ibv_wc *> *taskqueue, std::unordered_map<uint32_t, RdmaSalEndpoint *> *qpSalEndpointMap)
    : _id(id), _qpSalEndpointMap(qpSalEndpointMap)
{
    _taskQueue = taskqueue;
    if (pthread_create(&thread, NULL, &TaskThread::run, this))
    {
        std::cout << "pthread create has been failed while creating taskthread " << std::endl;
        exit(0);
    }
    cpu_set_t cpuset;
    CPU_ZERO(&cpuset);
    CPU_SET(cpu, &cpuset);
    if (pthread_setaffinity_np(thread, sizeof(cpu_set_t), &cpuset) != 0)
    {
        std::cerr << "Error calling pthread_setaffinity_np: "
                  << "\n";
    }
}

TaskThread::TaskThread(int id, ConcurrentQueue<struct ibv_wc *> *taskqueue, std::unordered_map<uint32_t, RdmaSalEndpoint *> *qpSalEndpointMap)
    : _id(id), _qpSalEndpointMap(qpSalEndpointMap)
{
    _taskQueue = taskqueue;
    if (pthread_create(&thread, NULL, &TaskThread::run, this))
    {
        std::cout << "pthread create has been failed while creating taskthread " << std::endl;
        exit(0);
    }
}

TaskThread::~TaskThread()
{
    std::cout << "Task Destructed" << std::endl;
    stop();
}
void TaskThread::stop()
{
    _stop = true;
    if (pthread_join(thread, NULL) == 0)
    {
        std::cout << "pthread join failed" << std::endl;
    }
}
inline void *TaskThread::run(void *object)
{
    TaskThread *thread = reinterpret_cast<TaskThread *>(object);
    std::cout << "running task thread" << thread->_id << std::endl;
    while (!thread->_stop)
    {
        struct ibv_wc *data = NULL;
        thread->_taskQueue->wait_and_pop(data);
        thread->processEvent(data);
    }
    return NULL;
}

void TaskThread::processEvent(struct ibv_wc *data)
{
    if (data == NULL || data->status != IBV_WC_SUCCESS)
    {
        std::cout << "TaskThread : failed work completion : ";
        std::cout << ibv_wc_status_str(data->status) << " on qp " << data->qp_num << std::endl;
        return;
    }
    /*
     * Process Request from client
     */

    auto it = _qpSalEndpointMap->find(data->qp_num);
    if (it == _qpSalEndpointMap->end())
    {
        std::cout << data->qp_num << "RdmaSal : endpoint not registered for qp num\n";
        return;
    }
    switch (data->opcode)
    {
    case IBV_WC_SEND:
        it->second->processSendCompletion(data);
        break;
    case IBV_WC_RECV:
    {
        // it->second->processRecvCompletion(data);
        char *buffer = new char[data->byte_len];
        memcpy(buffer, (void *)data->wr_id, data->byte_len);
        rdma_post_recv(it->second->_cm_id, (void *)data->wr_id, (void *)data->wr_id,
                       it->second->_recvMsgSize, it->second->_recvMr);
        struct SalRequestHeader *req = (struct SalRequestHeader *)buffer;
        std::cout << "recieve id : " << req->id << " " << (char *)req + SalRequestHeaderSize;
        std::cout << " " << req->type << "size" << data->byte_len << "\n";
        switch (req->type)
        {
        case RequestType::GET:
            it->second->processGet(req);
            break;
        case RequestType::DELETE:
            std::cout<<"TaskThread:: incorrect delete request from client to follower"<<std::endl;
            break;
        case RequestType::PUT:
            std::cout<<"TaskThread:: incorrect put request from client to follower"<<std::endl;
            break;
        default:
            std::cout << "SalRequest invalid req type";
            break;
        }
        delete[] buffer;
    }
    break;
    case IBV_WC_RDMA_WRITE:
        std::cout << "rdma write completion\n";
        break;
    case IBV_WC_RDMA_READ:
        std::cout << "rdma read completion\n";
        break;
    default:
        std::cout << "TaskThread default opcode : " << data->opcode << std::endl;
        break;
    }
}